import 'dart:io';
import 'dart:async';

class RecursiveWatcher {
  Directory directory;
  bool followLinks;
  int events;

  RecursiveWatcher(this.directory);

  /// Starts to watch a directory recursively.
  /// It will recursively review directory each time, when one of `directoryEventsForReview` events will happen on any object with type of `Directory`.
  /// Use `Directory(path).recursiveWatcher()` to construct `RecursiveWatcher`'s object.
  Stream<FileSystemEvent> watch(
      {bool followLinks = false,
      int events = FileSystemEvent.all,
      int directoryEventsForReview =
          FileSystemEvent.create | FileSystemEvent.move}) async* {
    var repeat = false;

    do {
      repeat = false;
      var stream = await directory._watchRecursively(
          followLinks: followLinks, events: events);

      await for (var event in stream) {
        yield event;
        if (event.isDirectory && (directoryEventsForReview & event.type) != 0) {
          repeat = true;
          break;
        }
      }
    } while (repeat);
  }
}

extension WatchingRecursively on Directory {
  /// Returns `RecursiveWatcher` which has `watch` method.
  /// Sample: 
  /// ```
  /// var stream = await Directory('.').recursiveWatcher().watch();
  /// stream.listen((event) => print("${event.path} ${event.isDirectory} ${event.type}"));
  /// ```
  RecursiveWatcher recursiveWatcher() => RecursiveWatcher(Directory(this.path));

  Future<Stream<FileSystemEvent>> _watchRecursively(
      {int events = FileSystemEvent.all, bool followLinks = true}) async {
    var controller = StreamController<FileSystemEvent>();

    watch(events: events, recursive: false).listen((e) => controller.add(e));

    var dirs = list(recursive: true, followLinks: followLinks)
        .where((e) => e is Directory);

    await for (var dir in dirs) {
      dir.watch().listen((e) => controller.add(e));
    }

    return controller.stream;
  }
}
